;------------------------------------------------------------------------------
; Title:			ds30 loader for dsPIC33F
;
; File description:	Main firmwarefile
;
; Copyright: 		Copyright  09-10, Mikael Gustafsson
;
; Version			1.1.0 february 2010
;
; Webpage: 			http://mrmackey.no-ip.org/elektronik/ds30loader/
;
; History:			1.1.0 New feature: bootloader protection
;						  New feature: auto baudrate detection
;						  New option: high baud rates selection
;					1.0.4 Bugfix: rs485 not working, nop inserted before check of trmt in Send()
;					      New feature: 0x00 goto protection
;					1.0.3 Added tx enable support, new Send function
;					1.0.2 Erase is now made just before write to increase reliability
;					1.0.1 Fixed baudrate error check
;					1.0.0 Added flash verification
;						  Corrected buffer variable location to .bss
;						  Buffer is now correctly sized
;					0.9.4 Fixed config write, byte instead of word
;					0.9.3 Added code to restore pps settings
;						  Removed initialization of stack limit register
;						  BRG is rounded instead of truncated
;					0.9.2 Removed PLL and clockswitching code
;					0.9.1 Added pps code
;						  Added baudrate error check
;						  Added clockswitching for frc+ppl
;					0.9.0 First version released, based on the dsPIC30F version
;                                          
;------------------------------------------------------------------------------

;-----------------------------------------------------------------------------
;    This file is part of ds30 Loader.
;
;    ds30 Loader is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation.
;
;    ds30 Loader is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with ds30 Loader. If not, see <http://www.gnu.org/licenses/>.
;------------------------------------------------------------------------------ 

;------------------------------------------------------------------------------
;
; USAGE  USAGE  USAGE  USAGE  USAGE  USAGE  USAGE  USAGE  USAGE  USAGE  USAGE
;
;------------------------------------------------------------------------------
;
; This MPLAB-project comes preconfigured for the dsPIC33FJ128GP204 device using
; internal FRC without pll and uart1
;
; To use other chip and/or configuration you need to do:
; * Select correct device in MPLAB IDE, menu Configure->Select Device
; * Include correct linker script
; * Search for xxx in settings.inc and check/alter those lines
; * If your device has the PPS(peripheral pin select) feature you need to configure that, template below, search for xxx
; * Put your own code in this file at designated areas, like i/o initialization etc. (optional)
; * No device specific errata workarounds are implemented in the code below
;
; Tabsize used is 4
;
;------------------------------------------------------------------------------


;------------------------------------------------------------------------------
; Register usage
;------------------------------------------------------------------------------
		;.equ	MIXED,		W0		;immediate
		;.equ	MIXED,		W1		;immediate
		.equ	WBUFPTR,	W2		;buffer pointer
		.equ	WCNT,		W3		;loop counter
		.equ	WADDR2,		W4		;memory pointer
		.equ	WADDR,		W5		;memory pointer
		.equ	WPPSTEMP1,	W6		;used to restore pps register
		.equ	WPPSTEMP2,	W7		;used to restore pps register
		.equ	WTEMP1,		W8		;
		.equ	WTEMP2,		W9		;
		.equ	WDEL1,		W10		;delay outer
		.equ	WDEL2,		W11		;delay inner
		.equ	WDOERASE,	W12		;flag indicated erase should be done before next write
		.equ	WCMD,		W13		;command
		.equ 	WCRC, 		W14		;checksum
		.equ	WSTPTR,		W15		;stack pointer


;------------------------------------------------------------------------------
; Includes
;------------------------------------------------------------------------------
		.include "settings.inc"	


;-----------------------------------------------------------------------------
; UART stuff
;------------------------------------------------------------------------------ 
		.ifndef USE_UART1
			.ifndef USE_UART2
				.error "No uart is specified"
			.endif
		.endif
		
		.ifdef USE_UART1
			.ifdef USE_UART2
				.error "Both uarts are specified"
			.endif

		   	.equ    UMODE,	    U1MODE					;uart mode
		   	.equ    USTA,  		U1STA					;uart status
		   	.equ    UBRG,		U1BRG					;uart baudrate
		   	.equ    UTXREG,		U1TXREG					;uart transmit
		   	.equ	URXREG,		U1RXREG					;uart receive
			.equ	UIFS,		IFS0					;uart interupt flag sfr
			.equ	URXIF,		U1RXIF					;uart received interupt flag
			.equ	UTXIF,		U1TXIF					;uart transmit interupt flag		   	
		.endif
		
		.ifdef USE_UART2
			.ifndef HAS_UART2
				.error "UART2 specified for a device that only has uart1"
			.endif		
		
		   	.equ    UMODE,	    U2MODE					;uart mode
		   	.equ    USTA,  		U2STA					;uart status
		   	.equ    UBRG,		U2BRG					;uart baudrate
		   	.equ    UTXREG,		U2TXREG					;uart transmit
		   	.equ	URXREG,		U2RXREG					;uart receive
			.equ	UIFS,		IFS1					;uart interupt flag sfr
			.equ	URXIF,		U2RXIF					;uart received interupt flag
			.equ	UTXIF,		U2TXIF					;uart transmit interupt flag		   	
		.endif


;------------------------------------------------------------------------------
; Constants
;------------------------------------------------------------------------------
		.equ	VERMAJ,		1										/*firmware version major*/
		.equ	VERMIN,		1										/*fimrware version minor*/
		.equ	VERREV,		0										/*firmware version revision*/

		.equ 	HELLO, 		0xC1		
		.equ 	OK, 		'K'										/*erase/write ok*/
		.equ 	CHECKSUMERR,'N'										/*checksum error*/
		.equ	VERFAIL,	'V'										/*verification failed*/
		.equ   	BLPROT,     'P'                              		/*bl protection tripped*/		
		.equ   	UCMD,     	'U'                              		/*unknown command*/		
		

		.equ	BLDELAY,	( BLTIME * (FCY / 1000) / (65536 * 7) )	/*delay before user application is loaded*/
		.ifdef USE_BRGH
			.equ	UARTBR,		( (((FCY / BAUDRATE) / 2) - 1) / 2 )	/*brg calculation with rounding*/
		.else
			.equ	UARTBR,		( (((FCY / BAUDRATE) / 8) - 1) / 2 )	/*brg calculation with rounding*/
		.endif

		.equ	BLPLP,		1										/*bootloader placement, pages from end*/
		.equ	PAGESIZER,	8										/*pagesize [rows]*/
		.equ	ROWSIZEW,	64										/*rowsize [words]*/		
		.equ	STARTADDR,	( FLASHSIZE - BLPLP * PAGESIZER * ROWSIZEW * 2 )	/*bootloader placement*/		
		.equ	BLSTARTROW,	(STARTADDR / ROWSIZEW / 2)	
		.equ	BLENDROW,	(STARTADDR / ROWSIZEW / 2 + PAGESIZER - 1)	
		

;------------------------------------------------------------------------------
; Validate user settings
;------------------------------------------------------------------------------
		; Internal cycle clock
		.if FCY > 40000000
			.error "Fcy specified is out of range"
		.endif

		; Baudrate error
		.ifdef USE_BRGH
			.equ REALBR,	( FCY / (4 * (UARTBR+1)) )
		.else
			.equ REALBR,	( FCY / (16 * (UARTBR+1)) )
		.endif
		.equ BAUDERR,	( (1000 * ( BAUDRATE - REALBR)) / BAUDRATE )
		.if ( BAUDERR > 25) || (BAUDERR < -25 )
			.error "Baudrate error is more than 2.5%. Remove this check or try another baudrate and/or clockspeed."
		.endif 
		

;------------------------------------------------------------------------------
; Global declarations
;------------------------------------------------------------------------------
        .global __reset          	;the label for the first line of code, needed by the linker script


;------------------------------------------------------------------------------
; Uninitialized variables in data memory
;------------------------------------------------------------------------------
		.bss
buffer:	.space ( ROWSIZEW * 3 + 1/*checksum*/ ) 


;------------------------------------------------------------------------------
; Send macro
;------------------------------------------------------------------------------
		.macro SendL char
			mov #\char, W0
			rcall Send
		.endm


;------------------------------------------------------------------------------
; Start of code section in program memory
;------------------------------------------------------------------------------
		.section *, code, address(STARTADDR-4)
usrapp:	nopr						;these two instructions will be replaced
		nopr						;with a goto to the user application by the pc program
		

;------------------------------------------------------------------------------
; Reset vector
;------------------------------------------------------------------------------
		.section *, code, address(STARTADDR)
__reset:mov		#__SP_init, WSTPTR	;initalize the Stack Pointer
		
			
;------------------------------------------------------------------------------
; User specific entry code go here, see also user exit code section at end of file
;------------------------------------------------------------------------------
		
		;----------------------------------------------------------------------
		; UART pps config
		;----------------------------------------------------------------------
		.ifdef HAS_PPS
; Lubin			.error "UART pps is not configured. Read datasheet and configure pps."			;xxx remove this line			
			
			; ** IMPORTANT **
			;
			; THIS CODE IS JUST A TEMPLATE AND WILL MOST LIKELY NOT WORK FOR YOU, 
			; READ THE DATASHEET AND ALTER LINES MARKED WITH XXX
			;
			; ** IMPORTANT **
			
			; Backup, these are restored in exit code at end of file
			; Changes needs to be done in exit, search for xxx
			mov		RPINR18, WPPSTEMP1		;xxx		
			mov		RPOR1, WPPSTEMP2			;xxx

			; Receive, map pin (14 = 1110) to uart
;			bclr	RPINR18, #U1RXR0		;xxx
;			bset	RPINR18, #U1RXR1		;xxx
;			bset	RPINR18, #U1RXR2		;xxx
;			bset	RPINR18, #U1RXR3		;xxx
;			bclr	RPINR18, #U1RXR4		;xxx

			; configure PLL
			mov #30,W0;
			mov W0,PLLFBD;
			mov #192,W0;
			mov W0,CLKDIV;
			
;			; clock switch
;			;Place the new oscillator selection in W0
;			;OSCCONH (high byte) Unlock Sequence
;			MOV #3,W0;
;			MOV #OSCCONH, w1
;			MOV #0x78, w2
;			MOV #0x9A, w3
;			MOV.b w2, [w1]
;			MOV.b w3, [w1]
;			;Set new oscillator selection
;			MOV.b WREG, OSCCONH
;			;OSCCONL (low byte) unlock sequence
;			MOV #OSCCONL, w1
;			MOV #0x46, w2
;			MOV #0x57, w3
;			MOV.b w2, [w1]
;			MOV.b w3, [w1]
;			;Start oscillator switch


			mov 	#48,W0;			Lubin ??
			mov 	W0,AD1PCFGL ;
			mov 	#16375,W0;		//65470
			mov		W0,TRISB

			mov		#0x46,W0;		Lubin Unlock
			mov		W0,OSCCON;
			mov		#0x57,W0;
			mov		W0,OSCCON;
			bclr 	_OSCCON,#6;

			; Transmit, map uart to pin

			mov 	#65282,W0	;		Lubin 
			mov 	W0,RPINR18 ;
				
			mov 	#1023,W0;		Lubin	
			mov		W0,RPOR1;

			mov		#0x46,W0;		Lubin unlock
			mov		W0,OSCCON;
			mov		#0x57,W0;
			mov		W0,OSCCON;

			bset 	_OSCCON,#6;


		
		.endif	

      			
;------------------------------------------------------------------------------
; Init 
;------------------------------------------------------------------------------
		clr		WDOERASE
		
		;		
		;bset	TRISB, #TRISB14 		;make input (datasheet pg 157)
		;bset	AD1PCFGL, #PCFG9		;analog pin to digital
		;bset	AD1PCFGL, #PCFG10		;analog pin to digital
		
		; UART		
		.ifdef USE_BRGH	
			bset	UMODE, #BRGH
		.endif
		
		;----------------------------------------------------------------------
		;UART without auto baudrate detection
		;----------------------------------------------------------------------
		.ifndef USE_ABAUD
			mov		#UARTBR, W0 		;set	
			mov 	W0, UBRG			; baudrate
			bset 	UMODE, #UARTEN		;enable uart
			bset 	USTA, #UTXEN		;enable transmit			
			
			
		;----------------------------------------------------------------------
		;UART with auto baudrate detection
		;----------------------------------------------------------------------
		.else
			setm	UBRG			;errata workaround
			bset	UMODE, #ABAUD	;enable auto baudrate
			bset 	UMODE, #UARTEN	;enable uart			
			
			;------------------------------------------------------------------
			; Wait for auto baudrate to complete	
			;------------------------------------------------------------------
			; Init delay
			mov 	#BLDELAY, WDEL1
			; Check for auto baud complete
arpt1:		clr		WDEL2
arptc:		clrwdt
			btss 	UMODE, #ABAUD
			bra		abaudok
	 		; Delay
anotrcv:	dec 	WDEL2, WDEL2
			bra 	nz, arptc
			dec 	WDEL1, WDEL1
			bra 	nz, arpt1
			; If we get here, abaud timed out	
			bra		exit	
			; Auto baud rate detection completed
abaudok:	bset 	USTA, #UTXEN		;enable transmit			
			SendL	OK					;confirm baudrate detection to application			
		.endif
		
		
		;----------------------------------------------------------------------
		; TX enable, make tx enable pin output and set to 0
		;----------------------------------------------------------------------
		.ifdef USE_TXENABLE
			bclr	TRISR_TXE, #TRISB_TXE
			bclr	LATR_TXE, #LATB_TXE
		.endif


;------------------------------------------------------------------------------
; Receive hello
;------------------------------------------------------------------------------
test:
		;SendL 	DEVICEID
		;bra		test

;		SendL 	#0x55
;		rcall Receive
;		bra		test


		rcall 	Receive
		sub 	#HELLO, W0			;check
		bra 	nz, exit			; prompt


;------------------------------------------------------------------------------
; Send device id and firmware version
;------------------------------------------------------------------------------
		SendL 	DEVICEID
		SendL	VERMAJ
		SendL	(VERMIN*16 + VERREV)
		

;------------------------------------------------------------------------------
; Main
;------------------------------------------------------------------------------
		; Send ok
Main:	SendL 	OK

		; Init checksum
main1:	clr 	WCRC

	
		;----------------------------------------------------------------------
		; Receive address
		;----------------------------------------------------------------------
		; Upper byte
		rcall 	Receive
		mov 	W0, TBLPAG
		; High byte, use PR1 as temporary sfr
		rcall 	Receive		
		mov.b	WREG, PR1+1
		; Low byte, use PR1 as temporary sfr
		rcall 	Receive
		mov.b	WREG, PR1
		;
		mov		PR1, WREG
		mov		W0,	WADDR
		mov		W0, WADDR2
		
		
		;----------------------------------------------------------------------
		; Receive command
		;----------------------------------------------------------------------
		rcall 	Receive
		mov		W0, WCMD
		

		;----------------------------------------------------------------------
		; Receive nr of data bytes that will follow
		;----------------------------------------------------------------------
		rcall 	Receive				
		mov 	W0, WCNT
	

		;----------------------------------------------------------------------
		; Receive data		
		;----------------------------------------------------------------------
		mov 	#buffer, WBUFPTR
rcvdata:
		rcall 	Receive				
		mov.b 	W0, [WBUFPTR++]
		dec		WCNT, WCNT
		bra 	nz, rcvdata			;last byte received is checksum	
		

		;----------------------------------------------------------------------
		; 0x00 goto protection
		;----------------------------------------------------------------------	
		.ifdef	PROT_GOTO
			cp0		TBLPAG
			bra		nz, chksum
			cp0		PR1
			bra		nz, chksum
			
			;
			mov 	#buffer, WBUFPTR
			; 1st word upper byte = goto instruction
			mov.b 	#0x04, W0
			mov.b	W0, [WBUFPTR++] 		
			; 1st word  low byte = low address byte
			mov.b 	#(0xff & STARTADDR), W0
			mov.b 	W0, [WBUFPTR++] 
			; 1st word high byte = high address byte
			mov.b 	#(0xff & (STARTADDR>>8)), W0
			mov.b 	W0, [WBUFPTR++]	
			;2nd word upper byte = unused
			clr.b	[WBUFPTR++]	
			; 2nd word  low byte = upper address byte
			mov.b 	#(0xff & (STARTADDR>>16)), W0
			mov.b 	W0, [WBUFPTR++]  
			; 2nd word high byte = unused
			clr.b 	[WBUFPTR++]  
		.endif
			
				
		;----------------------------------------------------------------------
		; Check checksum
		;----------------------------------------------------------------------
chksum:	cp0.b 	WCRC
		bra 	z, ptrinit
		SendL 	CHECKSUMERR
		bra 	main1					
		
			
		;----------------------------------------------------------------------
		; Init pointer
		;----------------------------------------------------------------------			
ptrinit:mov 	#buffer, WBUFPTR
		
		
		;----------------------------------------------------------------------
		; Check command
		;----------------------------------------------------------------------			
		; Erase page
		btsc	WCMD,	#0		
		bra		eraseact		
		; Write config byte		0xF8 00 00 - 0xF8 00 0E
		btsc	WCMD, #3
		bra		config
		; Write row				0x00 02 00 - 0x02 AC 00 
		btsc	WCMD,	#1		
		bra		blprot
		; Else, unknown command
		SendL   UCMD		
		bra     main1
		
					
		;----------------------------------------------------------------------
		; "Erase activation"
		;----------------------------------------------------------------------		
eraseact:
		mov		#0xffff, WDOERASE
		bra		Main
					
					
		;----------------------------------------------------------------------
		; Bootloader protection
		;----------------------------------------------------------------------
blprot:	nop
		.ifdef PROT_BL
			; Calculate row number of received address
			mov		TBLPAG, W1
			mov		WADDR, W0
			mov		#(ROWSIZEW*2), WTEMP1
			repeat	#17
			div.ud	W0, WTEMP1;		W = received address / (rowsizew*2)
			; Received row < bl start row = OK
			mov		#BLSTARTROW, WTEMP1
			cp		W0, WTEMP1
			bra		N, blprotok
			; Received row > bl end row = OK
			mov		#BLENDROW, WTEMP1
			cp		WTEMP1, W0
			bra		N, blprotok		
			; Protection tripped
			SendL   BLPROT		
		    bra     main1
			; Restore WADDR2
blprotok:	mov		WADDR, WADDR2
		.endif
		
					
		;----------------------------------------------------------------------		
		; Erase page
		;----------------------------------------------------------------------		
		btss	WDOERASE, #0
		bra		program
		tblwtl	WADDR, [WADDR]		;"Set base address of erase block", equivalent to setting nvmadr/u in dsPIC30F?
		; Erase
		mov 	#0x4042, W0
		rcall 	Write	
		; Erase finished
		clr		WDOERASE
		
		
		;----------------------------------------------------------------------		
		; Write row
		;----------------------------------------------------------------------		
program:mov 	#ROWSIZEW, WCNT
		; Load latches
latlo:	tblwth.b 	[WBUFPTR++], [WADDR] 	;upper byte
		tblwtl.b	[WBUFPTR++], [WADDR++] 	;low byte
		tblwtl.b	[WBUFPTR++], [WADDR++] 	;high byte	
		dec 	WCNT, WCNT
		bra 	nz, latlo
		; Write flash row
		mov 	#0x4001, W0		
		rcall 	Write
		
		
		;----------------------------------------------------------------------		
		; Verify row
		;----------------------------------------------------------------------	
		mov 	#ROWSIZEW, WCNT
		mov 	#buffer, WBUFPTR	
		; Verify upper byte
verrow:	tblrdh.b [WADDR2], W0
		cp.b	W0, [WBUFPTR++]
		bra		NZ, vfail	
		; Verify low byte
		tblrdl.b [WADDR2++], W0
		cp.b	W0, [WBUFPTR++]
		bra		NZ, vfail
		; Verify high byte
		tblrdl.b [WADDR2++], W0
		cp.b	W0, [WBUFPTR++]
		bra		NZ, vfail
		;
		dec		WCNT, WCNT
		bra 	nz, verrow
		bra		Main
		
		
		;----------------------------------------------------------------------
		; Verify fail
		;----------------------------------------------------------------------
vfail:	SendL	VERFAIL
		bra		main1
		
				
		;----------------------------------------------------------------------
		; Write config byte
		;----------------------------------------------------------------------
		; Load latch
config:	tblwtl.b [WBUFPTR], [WADDR]			
		; Write config byte
		mov 	#0x4000, W0
		rcall 	Write
		; Write finished
		bra 	Main
			
		
;------------------------------------------------------------------------------
; Write()
;------------------------------------------------------------------------------
Write:	mov 	W0, NVMCON
		mov 	#0x55, W0
		mov 	W0, NVMKEY
		mov 	#0xAA, W0
		mov 	W0, NVMKEY
		bset 	NVMCON, #WR
		nop
		nop	
		; Wait for erase/write to finish	
compl:	btsc	NVMCON, #WR		
		bra 	compl				
		return

		
;------------------------------------------------------------------------------
; Send()
;------------------------------------------------------------------------------
Send:	; Enable tx
		.ifdef USE_TXENABLE
			bset	LATR_TXE, #LATB_TXE
			nop
		.endif		
		;Send byte
		mov 	WREG, UTXREG
		nop
		nop
		; Wait until transmit shift register is empty
txwait:	btss	USTA, #TRMT
		bra		txwait
		; Disable tx 
		.ifdef USE_TXENABLE
			bclr	LATR_TXE, #LATB_TXE
		.endif
		; Send complete
		return	
		
		
;------------------------------------------------------------------------------
; Receive()
;------------------------------------------------------------------------------
		; Init delay
Receive:mov 	#BLDELAY, WDEL1
		; Check for received byte
rpt1:	clr		WDEL2
rptc:	clrwdt						;clear watchdog
		btss 	USTA, #URXDA		
		bra 	notrcv
		mov 	URXREG, W0			
		add 	WCRC, W0, WCRC		;add to checksum
		return
 		; Delay
notrcv:	dec 	WDEL2, WDEL2
		bra 	nz, rptc
		dec 	WDEL1, WDEL1
		bra 	nz, rpt1
		; If we get here, uart receive timed out
        mov 	#__SP_init, WSTPTR	;reinitialize the Stack Pointer
        
		
;------------------------------------------------------------------------------
; Exit point, clean up and load user application
;------------------------------------------------------------------------------		
exit:	bclr	UIFS, #URXIF		;clear uart received interupt flag
		bclr	UIFS, #UTXIF		;clear uart transmit interupt flag
		bclr	USTA, #UTXEN		;disable uart transmit
		bclr 	UMODE, #UARTEN		;disable uart
		clr		PR1					;clear PR1, was used as temporary sfr
	
	
;------------------------------------------------------------------------------
; User specific exit code go here
;------------------------------------------------------------------------------
		.ifdef HAS_PPS
;			.error "PPS restoration is not configured."			;xxx remove this line
			mov		WPPSTEMP1, RPINR18	;xxx restore 
			mov		WPPSTEMP2, RPOR1		;xxx  pps settings
		.endif


;------------------------------------------------------------------------------
; Load user application
;------------------------------------------------------------------------------
        bra 	usrapp

	
;------------------------------------------------------------------------------
; End of code
;------------------------------------------------------------------------------
		.end

 
